# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import Tuple

import bpy


def new_node_material(name: str = "material"):
    material = bpy.data.materials.new(name=name)
    material.use_nodes = True
    return material


Color_T = Tuple[float, float, float, float]


def update_BSDF_node(
    material: bpy.types.Material,
    color: Color_T = None,
    subsurface: float = None,
    subsurfaceRadius: tuple = None,
    subsurfaceColor: Color_T = None,
    metallic: float = None,
    specular: float = None,
    specularTint: float = None,
    roughness: float = None,
    anisotropic: float = None,
    anisotropicRotation: float = None,
    sheen: float = None,
    sheenTint: float = None,
    clearcoat: float = None,
    clearcoatRoughness: float = None,
    IOR: float = None,
    transmission: float = None,
    transmissionRoughness: float = None,
    emission: Color_T = None,
    emissionStrength: float = None,
    alpha: float = None,
) -> None:
    """Updates default values in Principled BSDF node of material.
    None params are ignored and doesn't modify node.

    :param material: Material to modify.
    :type material: bpy.types.Material
    :param color: Diffuse or metal surface color.
    :type color: Color_T, optional
    :param subsurface: Mix between diffuse and subsurface scattering. Rather than being a simple mix between Diffuse and Subsurface Scattering, it acts as a multiplier for the Subsurface Radius.
    :type subsurface: float, optional
    :param subsurfaceRadius: Average distance that light scatters below the surface. Higher radius gives a softer appearance, as light bleeds into shadows and through the object. The scattering distance is specified separately for the RGB channels, to render materials such as skin where red light scatters deeper. The X, Y and Z values are mapped to the R, G and B values, respectively.
    :type subsurfaceRadius: tuple, optional
    :param subsurfaceColor: Subsurface scattering base color.
    :type subsurfaceColor: Color_T, optional
    :param metallic: Blends between a non-metallic and metallic material model. A value of 1.0 gives a fully specular reflection tinted with the base color, without diffuse reflection or transmission. At 0.0 the material consists of a diffuse or transmissive base layer, with a specular reflection layer on top.
    :type metallic: float, optional
    :param specular: Amount of dielectric specular reflection. Specifies facing (along normal) reflectivity in the most common 0 - 8% range.
    :type specular: float, optional
    :param specularTint: Tints the facing specular reflection using the base color, while glancing reflection remains white.
    :type specularTint: float, optional
    :param roughness: Specifies microfacet roughness of the surface for diffuse and specular reflection.
    :type roughness: float, optional
    :param anisotropic: Amount of anisotropy for specular reflection. Higher values give elongated highlights along the tangent direction; negative values give highlights shaped perpendicular to the tangent direction.
    :type anisotropic: float, optional
    :param anisotropicRotation: Rotates the direction of anisotropy, with 1.0 going full circle.
    :type anisotropicRotation: float, optional
    :param sheen: Amount of soft velvet like reflection near edges, for simulating materials such as cloth.
    :type sheen: float, optional
    :param sheenTint: Mix between white and using base color for sheen reflection.
    :type sheenTint: float, optional
    :param clearcoat: Extra white specular layer on top of others. This is useful for materials like car paint and the like.
    :type clearcoat: float, optional
    :param clearcoatRoughness: Roughness of clearcoat specular.
    :type clearcoatRoughness: float, optional
    :param IOR: Index of refraction for transmission.
    :type IOR: float, optional
    :param transmission: Mix between fully opaque surface at zero and fully glass like transmission at one.
    :type transmission: float, optional
    :param transmissionRoughness: With GGX distribution controls roughness used for transmitted light.
    :type transmissionRoughness: float, optional
    :param emission: Light emission from the surface, like the Emission shader.
    :type emission: Color_T, optional
    :param emissionStrength: Strength of the emitted light. A value of 1.0 will ensure that the object in the image has the exact same color as the Emission Color, i.e. make it ‘shadeless’.
    :type emissionStrength: float, optional
    :param alpha: Controls the transparency of the surface, with 1.0 fully opaque. Usually linked to the Alpha output of an Image Texture node.
    :type alpha: float, optional
    """
    BSDF_node = material.node_tree.nodes["Principled BSDF"]
    if color is not None:
        BSDF_node.inputs[0].default_value = color
    if subsurface is not None:
        BSDF_node.inputs[1].default_value = subsurface
    if subsurfaceRadius is not None:
        BSDF_node.inputs[2].default_value[0] = subsurfaceRadius[0]
        BSDF_node.inputs[2].default_value[1] = subsurfaceRadius[1]
        BSDF_node.inputs[2].default_value[2] = subsurfaceRadius[2]
    if subsurfaceColor is not None:
        BSDF_node.inputs[3].default_value = color
    if metallic is not None:
        BSDF_node.inputs[4].default_value = metallic
    if specular is not None:
        BSDF_node.inputs[5].default_value = specular
    if specularTint is not None:
        BSDF_node.inputs[6].default_value = specularTint
    if roughness is not None:
        BSDF_node.inputs[7].default_value = roughness
    if anisotropic is not None:
        BSDF_node.inputs[8].default_value = anisotropic
    if anisotropicRotation is not None:
        BSDF_node.inputs[9].default_value = anisotropicRotation
    if sheen is not None:
        BSDF_node.inputs[10].default_value = sheen
    if sheenTint is not None:
        BSDF_node.inputs[11].default_value = sheenTint
    if clearcoat is not None:
        BSDF_node.inputs[12].default_value = clearcoat
    if clearcoatRoughness is not None:
        BSDF_node.inputs[13].default_value = clearcoatRoughness
    if IOR is not None:
        BSDF_node.inputs[14].default_value = IOR
    if transmission is not None:
        BSDF_node.inputs[15].default_value = transmission
    if transmissionRoughness is not None:
        BSDF_node.inputs[16].default_value = transmissionRoughness
    if emission is not None:
        BSDF_node.inputs[17].default_value = color
    if emissionStrength is not None:
        BSDF_node.inputs[18].default_value = emissionStrength
    if alpha is not None:
        BSDF_node.inputs[19].default_value = alpha


def set_material(ob: bpy.types.Object, material: bpy.types.Material):
    if len(ob.data.materials) == 0:
        ob.data.materials.append(material)
    else:
        ob.data.materials[0] = material



# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import Callable
from typing import Iterable
from typing import List
from typing import Tuple

import bmesh
import bpy
from bmesh.types import BMEdge
from bmesh.types import BMFace
from bmesh.types import BMVert
from mathutils import Vector

from PyR3.shortcut.context import Objects


class OperationCancelled(Exception):
    pass


def manual_set_edit_mode():
    bpy.ops.object.mode_set(mode="EDIT")


def manual_set_object_mode():
    bpy.ops.object.mode_set(mode="OBJECT")


class Edit:
    """class for automatic in-out switching Edit mode.
    It is meant to be used as context manager with edited
    object being passed as param to constructor.
    """

    _is_edit_mode: bool = False
    BMESH: bmesh.bmesh.types.BMesh = None
    ob: bpy.types.Object

    class MeshCompList(list):
        def selected(self) -> Edit.MeshCompList:
            return Edit.MeshCompList(element for element in self if element.select)

    def __init__(
        self,
        ob: bpy.types.Object,
        *more: bpy.types.Object,
        active: bpy.types.Object = None
    ) -> None:
        self.ob = ob
        self.active = active
        self.more = more

    @staticmethod
    def isEditMode():
        return bpy.context.object.mode == "EDIT"

    def __enter__(self) -> Edit:
        """Enters edit mode, selects everything in there."""
        if self.active is None:
            Objects.active = self.ob
        else:
            Objects.active = self.active
        Objects.select_only(self.ob)
        Objects.select(*self.more)
        manual_set_edit_mode()
        Edit._is_edit_mode = True
        self.BMESH = bmesh.from_edit_mesh(self.ob.data)
        self.select_all()
        return self

    def __exit__(self, class_, instance, traceback) -> None:
        """Return to object mode."""
        manual_set_object_mode()
        Edit._is_edit_mode = False

    def faces(self) -> MeshCompList[BMFace]:
        """Provides access to edited object bmesh attribute
        holding reference to list of all faces of edited mesh.

        :return: List of faces.
        :rtype: MeshCompList[BMFace]
        """
        self.BMESH.faces.ensure_lookup_table()
        return Edit.MeshCompList(self.BMESH.faces)

    def edges(self) -> MeshCompList[BMEdge]:
        """Provides access to edited object bmesh attribute
        holding reference to list of all edges of edited mesh.

        :return: List of edges.
        :rtype: MeshCompList[BMEdge]
        """
        self.BMESH.faces.ensure_lookup_table()
        return Edit.MeshCompList(self.BMESH.edges)

    def vertices(self) -> MeshCompList[BMVert]:
        """Access to edited object bmesh vertice table.

        :return: Vertices
        :rtype: MeshCompList[BMVert]
        """
        self.BMESH.verts.ensure_lookup_table()
        return Edit.MeshCompList(self.BMESH.verts)

    def get_selected_vertices(self) -> List[BMVert]:
        return [v for v in self.vertices() if v.select]

    def select_vertices(
        self,
        condition: Callable[[Vector], bool],
    ):
        """Selects vertices, when condition function returns true.

        :param condition: test callable. It will be given vertice coordinate as parameter.
        :type condition: Callable[Vector], bool]
        """
        for v in self.vertices():
            if condition(v.co):
                v.select = True

    def select_edges(
        self,
        condition: Callable[[Vector, Vector], bool],
    ):
        """Selects edges, when condition function returns true.

        :param condition: Test callable. It will be given edge vertice coordinate as parameter.
        :type condition: Callable[[Vector, Vector], bool]
        """
        for e in self.edges():
            if condition(e.verts[0].co, e.verts[1].co):
                e.select = True

    def select_facing(self, direction: Vector) -> Edit:
        if not isinstance(direction, Vector):
            direction = Vector(direction)
        for face in self.faces():
            if face.normal.dot(direction) == 1:
                face.select = True
        return self

    def select_all(self):
        """Selects whole mesh"""
        bpy.ops.mesh.select_all(action="SELECT")

    def deselect_all(self):
        """Deselects whole mesh"""
        bpy.ops.mesh.select_all(action="DESELECT")

    def invert_selection(self):
        """Invertices selection of mesh components."""
        bpy.ops.mesh.select_all(action="INVERT")

    def delete_vertices(self):
        """Delete selected vertices."""
        bpy.ops.mesh.delete(type="VERT")

    def delete_edges(self):
        """Delete selected edges."""
        bpy.ops.mesh.delete(type="EDGE")

    def delete_faces(self):
        """Delete selected faces."""
        bpy.ops.mesh.delete(type="FACE")

    def duplicate(self, mode: int = 1):
        """Duplicate selected."""
        bpy.ops.mesh.duplicate(mode=mode)

    # methods from blender API, moved here for easier access
    bevel = bpy.ops.mesh.bevel
    extrude = bpy.ops.mesh.extrude_region
    extrude_repeat = bpy.ops.mesh.extrude_repeat
    extrude_individual_faces = bpy.ops.mesh.extrude_edges_indiv
    edge_face_add = bpy.ops.mesh.edge_face_add
    collapse = bpy.ops.mesh.edge_collapse
    remove_doubles = bpy.ops.mesh.remove_doubles
    normals_make_consistent = bpy.ops.mesh.normals_make_consistent



# -*- coding: utf-8 -*-
from __future__ import annotations

import math
import re
from numbers import Number
from typing import List
from typing import Tuple

from .Field import Field


class _SuffixParser:

    _float_regex = r"(?P<VALUE>[\-+]?[0-9]*\.?[0-9]+)"
    tokens: List[re.Pattern, float]
    suffixes: List[str]

    def __init__(self, suffix_to_value: Tuple[Tuple[str, float]]) -> None:
        self.tokens = []
        self.suffixes = []
        for suffix, multiplier in suffix_to_value:
            self.suffixes.append(suffix)
            token = re.compile(f"{self._float_regex}{suffix}")
            self.tokens.append((token, multiplier))

    def parse(self, string: str) -> float:
        total = 0
        index = 0
        while index < len(string):
            for token, multiplier in self.tokens:
                token: re.Pattern
                if match := token.match(string, pos=index):
                    total += float(match.groupdict()["VALUE"]) * multiplier
                    index = match.end()
                    break
            else:
                index += 1
        return total

    def __str__(self) -> str:
        return f"SuffixParser, suffixes: {self.suffixes}"

    __repr__ = __str__


class Length(Field):
    """Accepts float with optional length unit suffix. Unit suffix causes
    float value to be converted to value with unit denoted by `output_unit`.

    Valid unit suffixes are:

        - **mil**  for mils

        - **in**   for inches

        - **ft**   for feets

        - **mm**   for millimeters

        - **cm**   for centimeters

        - **dm**   for decimeters

        - **m**    for meters

    Signs that doesn't match anything are ignored and treated as separators.
    """

    _suffix_to_value_map = (
        ("mil", 2.54 * 1e-5),
        ("in", 0.0254),
        ("ft", 0.3048),
        ("mm", 0.001),
        ("cm", 0.01),
        ("dm", 0.1),
        ("m", 1),
        ("", 1),
    )

    parser = _SuffixParser(_suffix_to_value_map)
    _suffix_to_value_map = dict(_suffix_to_value_map)

    def __init__(
        self,
        *,
        output_unit: str = "m",
        default: str | Number = None,
    ) -> None:
        self.output_divider = self._suffix_to_value_map[output_unit]
        if default is not None:
            self.default = self._digest_value(default)
        else:
            self.default = None

    def _digest_value(self, value: str | Number) -> float:
        if isinstance(value, str):
            return self.parser.parse(value)
        elif isinstance(value, Number):
            return float(value)
        else:
            self._raise_invalid_value_type(value)

    def digest(self, literal: str | Number = None) -> float:
        """Returns total value contained in the literal in meters.

        :param literal: literal to consume or Number
        :type literal: Union[str, Number]
        :raises TypeError: If other type than str or Number is given.
        :raises KeyError: If value is None and no default is given.
        :return: total in meters.
        :rtype: float
        """
        if literal is None:
            value = self._get_default()
        else:
            value = self._digest_value(literal)
        return self._convert_to_output_unit(value)

    def _convert_to_output_unit(self, value: float) -> float:
        return value / self.output_divider


class Angle(Length):
    """Accepts float with optional angle unit suffix. Unit suffix causes
    float value to be converted to value with unit denoted by `output_unit`.

    Valid unit suffixes are:

        - **rad** for radians

        - **π** / **pi** for radians, multiplied by π (3.14...)

        - **deg** for degrees

        - **"** / **sec**  for seconds of angle

        - **'** / **min**  for minutes of angle

    Signs that doesn't match anything are ignored and treated as separators.
    """

    _suffix_to_value_map = (
        ('"', math.pi / (3600 * 180)),
        ("sec", math.pi / (3600 * 180)),
        ("'", math.pi / (60 * 180)),
        ("min", math.pi / (60 * 180)),
        ("°", math.pi / 180),
        ("deg", math.pi / 180),
        ("π", math.pi),
        ("pi", math.pi),
        ("rad", 1),
        ("", 1),
    )

    parser = _SuffixParser(_suffix_to_value_map)
    _suffix_to_value_map = dict(_suffix_to_value_map)

    def __init__(
        self, *, output_unit: str = "rad", default: str | Number = None
    ) -> None:
        super().__init__(output_unit=output_unit, default=default)



# -*- coding: utf-8 -*-
from abc import ABC
from abc import abstractmethod
from typing import Any


class Field(ABC):

    _field_name: str = "Unknown"
    _factory_name: str = "Unknown"

    @abstractmethod
    def __init__(self, **kwargs) -> None:
        ...

    @abstractmethod
    def digest(self, value: Any = None) -> None:
        ...

    def _raise_missing_factory_field(self):
        raise KeyError(
            f"Missing Factory Field parameter for x{self._trace_location()}."
        )

    def _raise_invalid_value_type(self, value: Any = None):
        if value is None:
            raise TypeError(f"Value of invalid type given to {self._trace_location()}.")
        else:
            raise TypeError(
                f"Value '{value}', type '{type(value)}' is not valid in {self._trace_location()}."
            )

    def _set_trace_info(self, _field_name: str, _factory_name: str):
        self._field_name = _field_name
        self._factory_name = _factory_name

    def _get_default(self):
        if self.default:
            return self.default
        else:
            self._raise_missing_factory_field()

    def _trace_location(self):
        return f"{self._factory_name}::{self._field_name}"

    def __str__(self):
        return f"{self.__class__.__qualname__} Field"

    __repr__ = __str__



# -*- coding: utf-8 -*-
from __future__ import annotations

import math
from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from operator import getitem
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple

import bpy

from PyR3.shortcut.context import Objects


class _Modifier(ABC):
    def apply(self, **extra_params):
        MODIFIER = self.master_object.modifiers.new(
            self.__class__.__qualname__, self._get_modifier_type()
        )
        self._set_modifier_params(MODIFIER, self.__class__.__dataclass_fields__.keys())
        if extra_params:
            self._set_modifier_extra_params(MODIFIER, extra_params)
        self._apply_modifier(MODIFIER.name, self.master_object)

    @abstractmethod
    def _get_modifier_type(self) -> str:
        ...

    def _set_modifier_params(self, modifier: bpy.types.Modifier, param_keys: List[str]):
        for param_name in param_keys:
            if hasattr(modifier, param_name):
                param_value = getattr(self, param_name)
                setattr(modifier, param_name, param_value)

    def _set_modifier_extra_params(
        self, modifier: bpy.types.Modifier, extra: Dict[str, Any]
    ):
        for param_name in extra.keys():
            if hasattr(modifier, param_name):
                param_value = getitem(extra, param_name)
                setattr(modifier, param_name, param_value)

    def _apply_modifier(
        self, modifier_name: bpy.types.Modifier, ob: bpy.types.Object
    ) -> None:
        Objects.select_only(ob)
        Objects.active = ob
        bpy.ops.object.modifier_apply(modifier=modifier_name)


@dataclass
class Boolean(_Modifier):
    """Boolean Modifier wrapper. For documentation over modifier parameters
    visit https://docs.blender.org/api/current/bpy.types.BooleanModifier.html
    """

    master_object: bpy.types.Object
    object: bpy.types.Object
    operation: str = "DIFFERENCE"
    solver: str = "EXACT"
    use_self: bool = False

    def _get_modifier_type(self):
        return "BOOLEAN"


@dataclass
class Array(_Modifier):
    """Array modifier wrapper. For documentation over modifier parameters
    visit https://docs.blender.org/api/current/bpy.types.ArrayModifier.html
    """

    master_object: bpy.types.Object
    constant_offset_displace: Tuple[float, float, float] = (0, 0, 0)
    count: int = 1
    use_relative_offset: bool = False
    use_constant_offset: bool = True

    def _get_modifier_type(self):
        return "ARRAY"


@dataclass
class Solidify(_Modifier):
    """Solidify modifier wrapper. For documentation over modifier parameters
    visit https://docs.blender.org/api/current/bpy.types.SolidifyModifier.html
    """

    master_object: bpy.types.Object
    thickness: float = 0.01
    offset: float = -1
    use_even_offset: bool = False
    use_quality_normals: bool = True

    def _get_modifier_type(self):
        return "SOLIDIFY"


@dataclass
class Bevel(_Modifier):
    """Solidify modifier wrapper. For documentation over modifier parameters
    visit https://docs.blender.org/api/current/bpy.types.BevelModifier.html
    """

    master_object: bpy.types.Object
    affect: str = "EDGES"
    offset_type: str = "OFFSET"
    width: float = 0.1
    segments: int = 1
    limit_method: str = "NONE"
    angle_limit: float = math.pi / 6
    use_clamp_overlap: bool = True

    def _get_modifier_type(self):
        return "BEVEL"



"""IO module provides import/export functions with ability to recognize
file format from filename. It's limited but handy solution, hence available here.
Recognized formats are:

- **GLB** : glTF Binary (.glb), Exports a single file, with all data packed in binary form.

- **GLTF** : glTF Embedded (.gltf), Exports a single file, with all data packed in JSON.

- **FBX** : Autodesk Filmbox (.fbx)

- **X3D** : Extensible 3D Graphics (.x3d)

- **OBJ** : Wavefront OBJ (.obj)

- **PLY** : Polygon File Format / Polygon File Format (.ply)

- **STL** : STL triangle mesh data (.stl)

- **BLEND** / **BLEND1** : Blender file format (.blend/.blend1) be aware that it causes to overwrite current scene on import.
"""
# -*- coding: utf-8 -*-
from __future__ import annotations

from pathlib import Path as _Path

from bpy.ops import export_mesh as _export_mesh
from bpy.ops import export_scene as _export_scene
from bpy.ops import import_mesh as _import_mesh
from bpy.ops import import_scene as _import_scene
from bpy.ops import wm as _wm

__export = {
    "GLB": lambda filepath, **kwargs: _export_scene.gltf(
        filepath=filepath, export_format="GLB", **kwargs
    ),
    "GLTF": lambda filepath, **kwargs: _export_scene.gltf(
        filepath=filepath, export_format="GLTF_EMBEDDED", **kwargs
    ),
    "FBX": _export_scene.fbx,
    "X3D": _export_scene.x3d,
    "OBJ": _export_scene.obj,
    "PLY": _export_mesh.ply,
    "STL": _export_mesh.stl,
    "BLEND": _wm.save_as_mainfile,
    "BLEND1": _wm.save_as_mainfile,
}


def export_to(filepath: str, **kwargs):
    """Export all objects into file. Format is determined from file extension.
    kwargs will be forwarded to bpy method call coresponding to selected format.

    :param filepath: _Path to the file to export to.
    :type filepath: str
    :raises KeyError: if format is not recognized.
    """
    format = _Path(filepath).suffix[1:].upper()
    try:
        __export.get(format)(filepath=str(filepath), **kwargs)
    except KeyError:
        raise KeyError(f"Format {format} is not supported.") from None


__import = {
    "GLB": lambda filepath, **kwargs: _import_scene.gltf(
        filepath=filepath, import_format="GLB", **kwargs
    ),
    "GLTF": lambda filepath, **kwargs: _import_scene.gltf(
        filepath=filepath, import_format="GLTF_EMBEDDED", **kwargs
    ),
    "FBX": _import_scene.fbx,
    "X3D": _import_scene.x3d,
    "OBJ": _import_scene.obj,
    "PLY": _import_mesh.ply,
    "STL": _import_mesh.stl,
    "BLEND": _wm.open_mainfile,
    "BLEND1": _wm.open_mainfile,
}


def import_from(filepath: str, **kwargs):
    """Import data from file. Format is determined from file extension.
    kwargs will be forwarded to bpy method call coresponding to selected format.

    :param filepath: _Path to file to import.
    :type filepath: str
    :raises KeyError: if format is not recognized.
    """
    format = _Path(filepath).suffix[1:].upper()
    try:
        __import.get(format)(filepath=str(filepath), **kwargs)
    except KeyError:
        raise KeyError(f"Format {format} is not supported.") from None



# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import Tuple

import bpy


class Transform:

    """This class is a container for set of object transforming functions.
    They all operate on global (currently selected) object(s).
    """

    @staticmethod
    def apply(
        do_move: bool = False,
        do_rotation: bool = False,
        do_scale: bool = False,
    ):
        """Apply the object's transformation to its data.

        :param use_move: applies move if true, defaults to False
        :type use_move: bool, optional
        :param use_rotation: applies rotation if true, defaults to False
        :type use_rotation: bool, optional
        :param use_scale: applies scale if true, defaults to False
        :type use_scale: bool, optional
        """
        bpy.ops.object.transform_apply(
            location=do_move, rotation=do_rotation, scale=do_scale
        )

    @classmethod
    def move(
        cls,
        vector: Tuple[float, float, float],
        **kwargs,
    ):
        """Move selected objects.

        :param vector: absolute coordinates to move to
        :type vector: Tuple[float, float, float], optional
        :param `**kwargs`: All from `bpy.ops.transform.translate <https://docs.blender.org/api/current/bpy.ops.transform.html#bpy.ops.transform.translate>`_.
        """
        bpy.ops.transform.translate(
            value=vector,
            **kwargs,
        )

    @classmethod
    def rotate(
        cls,
        angle: float,
        orient_axis: str,
        **kwargs,
    ):
        """Rotate selected objects around `orient_axis`

        :param angle: rotation angle
        :type angle: float, optional
        :param orient_axis: axis to rotate around, either "X", "Y" or "Z".
        :type orient_axis: str, optional
        :param `**kwargs`: All from `bpy.ops.transform.rotate <https://docs.blender.org/api/current/bpy.ops.transform.html#bpy.ops.transform.rotate>`_.
        """
        bpy.ops.transform.rotate(
            angle=angle,
            orient_axis=orient_axis,
            **kwargs,
        )

    @classmethod
    def scale(
        cls,
        scales: Tuple[float, float, float],
        **kwargs,
    ):
        """Scale (resize) selected objects.

        :param scales: Tuple of scales for each axis, (x, y, z)
        :type scales: tuple, optional
        :param `**kwargs`: All from `bpy.ops.transform.resize <https://docs.blender.org/api/current/bpy.ops.transform.html#bpy.ops.transform.resize>`_.
        """
        bpy.ops.transform.resize(
            value=scales,
            **kwargs,
        )

    resize = scale



# -*- coding: utf-8 -*-
from __future__ import annotations

from collections import UserList
from contextlib import contextmanager
from typing import List

import bpy

Object = bpy.types.Object


class _ContextMeta(type):
    @property
    def active(cls) -> Object:
        return bpy.context.active_object

    @active.setter
    def active(cls, object_: Object):
        bpy.context.view_layer.objects.active = object_

    @property
    def selected(cls) -> Objects[Object]:
        return Objects(bpy.context.selected_objects)


class Objects(list, metaclass=_ContextMeta):

    """
    As a class, provides set of static functionalities for managing global
    selection and currently selected and active objects.

    As a instance, is a container for objects with possibility to select
    or deselect object(s) contained inside.
    """

    #: Currently active object. It can be set to change active object.
    active: Object
    #: List of currently selected objects.
    #: This property is read-only.
    #: Use methods to alter selection.
    selected: Objects[Object]

    @staticmethod
    def select(*ob: Object):
        """Select given object(s).

        :param ob: object(s) to select
        :type ob: Object
        """
        for ob_ in ob:
            ob_.select_set(True)

    @staticmethod
    def deselect(*ob: Object):
        """Deselect given object(s).

        :param ob: object(s) to deselect
        :type ob: Object
        """
        for ob_ in ob:
            ob_.select_set(False)

    @staticmethod
    def select_all():
        """Selects all objects available in viewport."""
        bpy.ops.object.select_all(action="SELECT")

    @staticmethod
    def deselect_all():
        """Deselects all objects available in viewport."""
        bpy.ops.object.select_all(action="DESELECT")

    @classmethod
    def select_only(cls, *ob: Object):
        """Deselects all objects and selects only given one(s).

        :param ob: Object(s) to select
        :type ob: Object
        """
        cls.deselect_all()
        cls.select(*ob)

    @classmethod
    def delete(cls, *ob: Object) -> None:
        """Delete object(s) ob. It keeps previously selected object(s) selected.

        :param ob: Object(s) to select
        :type ob: Object
        """
        cls.deselect(*ob)
        with temporarily_selected(*ob):
            bpy.ops.object.delete()

    @classmethod
    def delete_all(cls) -> None:
        """Deletes all selectable objects."""
        cls.select_all()
        bpy.ops.object.delete()

    @classmethod
    def duplicate(cls, *ob: Object) -> None:
        """Duplicates object(s)

        :param ob: object(s) to duplicate
        :type ob: Object
        """
        with temporarily_selected(*ob):
            bpy.ops.object.duplicate()

    @staticmethod
    def inverse_selection():
        bpy.ops.object.select_all(action="INVERT")

    def select_contained(self):
        """Selects elements contained in this sequence."""
        self.select(*self)

    def deselect_contained(self):
        """Deselects elements contained in this sequence."""
        self.deselect(*self)

    def select_only_contained(self):
        """Selects only elements contained in this sequence."""
        self.select_only(*self)

    def only(self) -> Object:
        """Returns only element in sequence.

        :raises ValueError: If sequence is empty or has more than one element.
        :return: Only element from sequence
        :rtype: Object
        """
        if len(self) == 1:
            return self[0]
        else:
            raise ValueError("Objects list doesn't contain one element.")

    def __str__(self) -> str:
        return f"Objects{super().__str__()}"

    @staticmethod
    def all():
        return Objects(bpy.context.scene.objects)


@contextmanager
def temporarily_selected(*ob: Object):
    """For context manager usage, on enter selects only objects
    passed to constructor, on exit restores selection on previously selected objects.
    """
    # preparation
    previously_selected = Objects.selected
    Objects.select_only(*ob)
    # yield execution to caller
    yield
    # back here for clean-up
    Objects.select_only(*previously_selected)


@contextmanager
def temporary_scene():
    """Creates temporary scene and sets it as currently used.
    After exit, all objects selected in temporary scene are
    copied into previous scene and previous scene is set as
    currently used.

    :yield: (new, old) scenes
    :rtype: Tuple[bpy.types.Scene, bpy.types.Scene]
    """
    # preperation
    old = getScene()
    newScene()
    new = getScene()
    # yield execution to caller
    yield new, old
    # clean-up code here
    for selected in Objects.selected:
        old.collection.objects.link(selected)
    for selected in Objects.selected:
        new.collection.objects.unlink(selected)
    delScene()


def getScene() -> bpy.types.Scene:
    """Returns currently used scene.

    :return: Scene
    :rtype: bpy.types.Scene
    """
    return bpy.context.window.scene


def setScene(scene: bpy.types.Scene) -> None:
    """Sets new scene to use.

    :param scene: Scene
    :type scene: bpy.types.Scene
    """
    bpy.context.window.scene = scene


def newScene() -> None:
    """Creates new Scene object and automatically
    sets it as currently used one."""
    bpy.ops.scene.new()


def delScene() -> None:
    """Deletes currently used scene."""
    bpy.ops.scene.delete()


def cleanScene() -> None:
    """Deletes current scene and creates new one.
    Be aware that for total clean-up you should call
    wipeScenes() instead, as it destroys **ALL** scenes,
    not only current one, as cleanScene() does.
    """
    old_scene = getScene()
    newScene()
    new_scene = getScene()
    setScene(old_scene)
    delScene()
    setScene(new_scene)


def listScenes() -> List[bpy.types.Scene]:
    """Returns list of existing scenes."""
    return list(bpy.data.scenes)


def wipeScenes() -> None:
    """Destroys all existing ones and creates new empty one."""
    for _ in listScenes()[:-1]:
        delScene()
    cleanScene()



# -*- coding: utf-8 -*-
from __future__ import annotations

import re

from .Field import Field


class String(Field):
    """String factory field. You can specify length and default for it.
    Its value is always a string.

    :param default: Default value, used if no value is provided.
        If None, exception will be raised if no value will be given, defaults to None
    :type default: str, optional
    :param min_length: Minimal length of string. Exception will be raised
        if requirement will not be fullfiled. Defaults to None
    :type min_length: int, optional
    :param max_length: Minimal length of string. Exception will be raised
        if requirement will not be fullfiled. Defaults to None
    :type max_length: int, optional
    """

    def __init__(
        self, *, default: str = None, min_length: int = None, max_length: int = None
    ) -> None:
        self.default = default
        self.min_length = min_length
        self.max_length = max_length

    def digest(self, value: str = None) -> str:
        """Consumes value and returns cleaned string.
        Raises exception if requirements for string
        format are not met.

        :param value: value to consume, defaults to None
        :type value: str, optional
        :return: cleaned string.
        :rtype: str
        """
        if value is None:
            return self._get_default()
        else:
            string = str(value)
            self._check_conditions(string)
            return string

    def _check_conditions(self, string: str):
        if self.min_length is not None:
            if not (self.min_length <= len(string)):
                raise ValueError(
                    f"String for {self._trace_location()} is too short ({len(string)}, min length: {self.min_length})"
                )
        if self.max_length is not None:
            if not (len(string) <= self.max_length):
                raise ValueError(
                    f"String for {self._trace_location()} is too long ({len(string)}, min length: {self.max_length})"
                )


class Regex(String):
    """String field with possibility to use regular expression
    to check if string format is valid.

    :param pattern: Regular expression patter. String will be automatically compiled.
    :type pattern: re.Pattern or str
    :param default: [description], defaults to None
    :type default: str, optional
    :param flags: regular expression flags, from re module, defaults to 0
    :type flags: int, optional
    :raises TypeError: if pattern is neither str or re.Pattern.
    """

    _pattern: re.Pattern

    def __init__(
        self, pattern: re.Pattern | str, *, default: str = None, flags: int = 0
    ) -> None:
        if isinstance(pattern, str):
            self._pattern = re.compile(pattern, flags=flags)
        elif isinstance(pattern, re.Pattern):
            self._pattern = pattern
        else:
            raise TypeError(f"Invalid type of pattern parameter: {type(pattern)}")
        self.default = default

    def _check_conditions(self, string: str):
        if not self._pattern.fullmatch(string):
            raise ValueError(
                f"Value '{string}' doesn't match pattern '{self._pattern.pattern}' of field {self._trace_location()}"
            )



# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import Any

from .Field import Field


class Select(Field):
    def __init__(self, *values, default_index: int = None) -> None:
        self.values = values
        self.default_index = default_index

    def digest(self, value: Any = None) -> None:
        if value is None:
            if self.default_index is not None:
                return self.values[self.default_index]
            else:
                self._raise_missing_factory_field()
        elif value in self.values:
            return value
        else:
            raise ValueError(
                f"Invalid value for {self._trace_location()}, must be one of {self.values}"
            )



"""Mesh operation shortcuts, including creation, bounding box calculations and more.
"""
# -*- coding: utf-8 -*-
from __future__ import annotations

import time
from typing import List
from typing import Tuple

import bpy
from bpy.ops import mesh
from bpy.types import Object
from mathutils import Vector

from PyR3.shortcut.context import Objects
from PyR3.shortcut.context import getScene


def __return_active(function):
    def call_and_return_active(*args, **kwargs):
        function(*args, **kwargs)
        return Objects.active

    return call_and_return_active


#: Shortcut for creating circle. It returns created object.
addCircle: mesh.primitive_circle_add = __return_active(mesh.primitive_circle_add)
#: Shortcut for creating cone. It returns created object.
addCone: mesh.primitive_cone_add = __return_active(mesh.primitive_cone_add)
#: Shortcut for creating uv sphere. It returns created object.
addUVSphere: mesh.primitive_uv_sphere_add = __return_active(
    mesh.primitive_uv_sphere_add
)
#: Shortcut for creating cube. It returns created object.
addCube: mesh.primitive_cube_add = __return_active(mesh.primitive_cube_add)
#: Shortcut for creating cylinder. It returns created object.
addCylinder: mesh.primitive_cylinder_add = __return_active(mesh.primitive_cylinder_add)
#: Shortcut for creating grid. It returns created object.
addGrid: mesh.primitive_grid_add = __return_active(mesh.primitive_grid_add)
#: Shortcut for creating ico sphere. It returns created object.
addIcoSphere: mesh.primitive_ico_sphere_add = __return_active(
    mesh.primitive_ico_sphere_add
)
#: Shortcut for creating plane. It returns created object.
addPlane: mesh.primitive_plane_add = __return_active(mesh.primitive_plane_add)
#: Shortcut for creating torus. It returns created object.
addTorus: mesh.primitive_torus_add = __return_active(mesh.primitive_torus_add)


def boundingBoxCenterPoint(ob: Object) -> Vector:
    """Calculates center of bounding box.

    :param ob: Object to calculate for.
    :type ob: Object
    :return: Center point.
    :rtype: Vector
    """
    local_bbox_center = 0.125 * sum((Vector(b) for b in ob.bound_box), Vector())
    global_bbox_center = ob.matrix_world @ local_bbox_center
    return global_bbox_center


def containingSphereRadius(ob: Object, center: Vector = None) -> float:
    """Calculate radius of a sphere that bounding box can fit in.

    :param ob: Object to calculate for
    :type ob: Object
    :param center: Changes center from bbox center, defaults to None
    :type center: Vector, optional
    :return: radius
    :rtype: float
    """
    if center is None:
        loc = Object.bboxCenter(ob)

    vector_max: Vector = max(
        [(ob.matrix_world @ Vector(bound)) - loc for bound in ob.bound_box]
    )
    return vector_max.length


def boundingBoxPoints(ob: Object) -> float:
    """Calculates object's bounding box

    :param bpy_obj: Object to get bbox of.
    :type bpy_obj: Object
    :return: List of bounding box points.
    :rtype: float
    """
    return [ob.matrix_world @ Vector(b) for b in ob.bound_box]


def convert(ob: Object, target: str = "MESH"):
    """Convert object from one type to another.

    :param ob: Object to transform
    :type ob: Object
    :param target: target object type, defaults to "MESH"
    :type target: str, optional
    """
    Objects.active = ob
    Objects.select_only(ob)
    bpy.ops.object.convert(target=target)


def join(target: Object, *rest: Object):
    """Joins rest objects into target object.
    This will result in merging meshes into one
    object's data.

    :param target: object to join rest into
    :type target: Object
    :param rest: other objects to join
    :type rest: Object
    """
    Objects.deselect_all()
    Objects.active = target
    Objects.select_only(target)
    Objects.select(*rest)
    bpy.ops.object.join()


def fromPyData(
    vertexData: List[Tuple[float, float, float]] = [],
    edgeData: List[Tuple[float, float]] = [],
    faceData: List[Tuple[float, ...]] = [],
    *,
    mesh_name="mesh",
    object_name="object",
) -> Object:
    """Creates new mesh object from python data.

    :param vertexData: list of vertices, defaults to []
    :type vertexData: List[Tuple[float, float, float]], optional
    :param edgeData: list of tuples of edges vertices indexes, defaults to []
    :type edgeData: List[Tuple[float, float]], optional
    :param faceData: list of tuples of faces edge indexes, defaults to []
    :type faceData: List[Tuple[float, ...]], optional
    :param mesh_name: name for mesh data object, defaults to "mesh"
    :type mesh_name: str, optional
    :param object_name: name for mesh object, defaults to "object"
    :type object_name: str, optional
    :return: created object.
    :rtype: Object
    """
    new_mesh = bpy.data.meshes.new(mesh_name)
    new_mesh.from_pydata(vertexData, edgeData, faceData)
    new_mesh.update()
    obj = bpy.data.objects.new(object_name, new_mesh)
    # getScene()
    bpy.context.scene.collection.objects.link(obj)
    return obj



# -*- coding: utf-8 -*-
from __future__ import annotations

from numbers import Number

from .Field import Field


class Integer(Field):
    def __init__(self, *, default: int = None, value_range: range = None) -> None:
        self.default = default
        self.value_range = value_range

    def digest(self, value: str | Number = None) -> None:
        if value is None:
            return self._get_default()
        else:
            parsed_int = int(value)
            self.check_if_in_range(parsed_int)
            return parsed_int

    def check_if_in_range(self, parsed_int):
        if self.value_range is not None:
            if parsed_int not in self.value_range:
                raise ValueError(
                    f"Value {parsed_int} out of desired value range in {self._trace_location()}."
                )


class Float(Field):
    def __init__(
        self, *, default: float = None, min: float = None, max: float = None
    ) -> None:
        self.default = default
        self.min = min
        self.max = max

    def digest(self, value: str | Number = None) -> None:
        if value is None:
            return self._get_default()
        else:
            parsed_float = float(value)
            self.check_if_in_range(parsed_float)
            return parsed_float

    def check_if_in_range(self, parsed_float):
        if self.min is not None:
            if not (self.min <= parsed_float):
                raise ValueError(
                    f"Value {parsed_float} below expected range. (min: {self.min}) in {self._trace_location()}"
                )
        if self.max is not None:
            if not (parsed_float <= self.max):
                raise ValueError(
                    f"Value {parsed_float} above expected range. (max: {self.max}) in {self._trace_location()}"
                )



# -*- coding: utf-8 -*-
import os
import platform
import shutil
import site
import tarfile
from pathlib import Path

import requests

IS_LINUX = platform.system() == "Linux"
IS_WINDOWS = platform.system() == "Windows"

BPY_WIN_DIST_NAME = "bpy_2_93-win_amd64-cp39.tar.gz"
BPY_LINUX_DIST_NAME = "bpy_2_93-linux_64-cp39.tar.gz"


def install_bpy_lib():
    if check_unpack_required():
        if IS_WINDOWS:
            unpack_windows()
        elif IS_LINUX:
            unpack_linux()
        else:
            raise RuntimeError(
                "This operating system is not supported. We support only Windows and Linux."
            )


def unpack_linux():
    tarfile_path = bpy_tar_gz_path()
    with tarfile.open(tarfile_path) as archive:
        archive.extractall(get_site_packages_dir(), members=archive.getmembers())
    os.remove(tarfile_path)


def unpack_windows():
    tarfile_path = bpy_tar_gz_path()
    with tarfile.open(tarfile_path) as archive:
        folder_2_93_members = [
            member
            for member in archive.getmembers()
            if member.path.startswith("./2.93/")
        ]
        archive.extractall(get_python_executable_dir(), members=folder_2_93_members)
        other_members = [
            member
            for member in archive.getmembers()
            if not member.path.startswith("./2.93/")
        ]
        archive.extractall(get_site_packages_dir(), members=other_members)
    os.remove(tarfile_path)


def check_unpack_required():
    try:
        import bpy
    except ImportError:
        return True
    return False


def is_installed():
    end_path = os.path.join("src", "PyR3", "unpack.py")
    if Path(__file__).__str__().endswith(end_path):
        return False
    return True


def get_python_executable_dir() -> Path:
    return Path(os.__file__).parent.parent


def get_site_packages_dir() -> Path:
    for path in site.getsitepackages():
        if "site-packages" in path:
            return Path(path)


if is_installed():

    def bpy_tar_gz_path() -> Path:
        lib_dir = Path(__file__).parent.parent.parent
        if IS_WINDOWS:
            path = lib_dir / BPY_WIN_DIST_NAME
            if not os.path.exists(path):
                download_bpy_binaries(
                    BPY_WIN_DIST_NAME,
                    path,
                )
        elif IS_LINUX:
            path = lib_dir / BPY_LINUX_DIST_NAME
            if not os.path.exists(path):
                download_bpy_binaries(
                    BPY_LINUX_DIST_NAME,
                    path,
                )
        return path


else:

    def bpy_tar_gz_path() -> Path:
        if IS_WINDOWS:
            path = (
                Path(__file__).parent.parent.parent / "lib" / "win" / BPY_WIN_DIST_NAME
            )
            if not os.path.exists(path):
                download_bpy_binaries(
                    BPY_WIN_DIST_NAME,
                    path,
                )
        elif IS_LINUX:
            path = (
                Path(__file__).parent.parent.parent
                / "lib"
                / "linux"
                / BPY_LINUX_DIST_NAME
            )
            if not os.path.exists(path):
                download_bpy_binaries(
                    BPY_LINUX_DIST_NAME,
                    path,
                )
        return path


def download_bpy_binaries(dist: str, local_path: Path):
    download_file(
        f"https://github.com/Argmaster/pyr3/releases/download/bpy-binaries/{dist}",
        local_path,
    )


def download_file(url: str, local_path: Path):
    local_path.parent.mkdir(parents=True, exist_ok=True)
    with requests.get(url, stream=True) as r:
        with open(local_path, "wb") as f:
            shutil.copyfileobj(r.raw, f)


def get_path_members(archive: tarfile.TarFile, path: str):
    return [member for member in archive.getmembers() if member.path.startswith(path)]


if __name__ == "__main__":
    install_bpy_lib()



# -*- coding: utf-8 -*-
from __future__ import annotations

from abc import ABCMeta
from abc import abstractmethod
from inspect import isclass
from typing import Tuple
from typing import Type

import bpy

from PyR3.factory.fields.Field import Field
from PyR3.shortcut.context import Objects
from PyR3.shortcut.context import temporary_scene


class _MeshFactoryMeta(ABCMeta):

    required_members = [
        ["__doc__", lambda v: isinstance(v, str)],
        ["__author__", lambda v: isinstance(v, str)],
        ["__version__", lambda v: isinstance(v, str)],
    ]

    def __new__(cls, name, bases, attributes) -> None:
        cls.check_has_members(name, attributes)
        attributes["__factory_fields__"] = cls.get_field_names(cls, name, attributes)
        attributes["__factory_fields__"] |= cls.get_inherited_fields(bases)
        instance = ABCMeta.__new__(cls, name, bases, attributes)
        cls.wrap_render(instance)
        return instance

    def get_field_names(cls, classname, attributes: dict) -> dict:
        fields = {}
        for name, field in attributes.items():
            if isinstance(field, Field):
                field._set_trace_info(name, classname)
                fields[name] = field
            elif isclass(field) and issubclass(field, Field):
                field = field()
                field._set_trace_info(name, classname)
                fields[name] = field
        return fields

    def get_inherited_fields(bases: Tuple[Type[MeshFactory]]) -> dict:
        inherited_fields = {}
        for base in bases:
            if issubclass(base, MeshFactory):
                inherited_fields.update(getfields(base))
        return inherited_fields

    @classmethod
    def check_has_members(cls, classname, attributes):
        for name, validator in cls.required_members:
            if (value := attributes.get(name, None)) is None:
                raise TypeError(
                    f"Trying to create class {classname} without required member {name}."
                )
            if not validator(value):
                raise TypeError(
                    f"Trying to create class {classname} with {name} of invalid type."
                )

    @classmethod
    def wrap_render(cls, instance: MeshFactory):
        old_render = instance.render

        def render(*args, **kwargs):
            with temporary_scene():
                old_render(*args, **kwargs)

        instance.render = render


class MeshFactory(metaclass=_MeshFactoryMeta):
    """
    Base class for a mesh factory object.
    Mesh factory requires __doc__, __author__ and __version__ to be defined
    in mesh factory subclass, otherwise class instantiation will fail.
    Mesh factory can (and should) make use of Fields (subclasses of Field class)
    to specify mesh factory customization params. See PyR3.factory.fields modules
    for first-party fields. To specify field just set class attribute to instance of
    Field subclass.
    See :doc:`MeshFactory usage <../usage/factory>`.
    """

    __author__ = "Krzysztof Wiśniewski"
    __version__ = "2.0.0"

    def __init__(self, params: dict) -> None:
        for name, field in getfields(self).items():
            param_value = params.get(name, None)
            cleaned_value = field.digest(param_value)
            setattr(self, name, cleaned_value)

    @abstractmethod
    def render(self):
        """Implements rendering process. You can access cleaned contents
        of fields via self.field_name. Rendering process shouldn't alter
        scenes, as it's always happening in isolated scene.
        Only things that are selected when this function returns will
        be copied into callers scene.
        """


def getfields(mesh_factory: MeshFactory) -> dict:
    """Returns fields specified for given MeshFactory.

    :param mesh_factory: Object to fetch fields from.
    :type mesh_factory: MeshFactory
    :return: dictionary of factory fields.
    :rtype: dict
    """
    if isclass(mesh_factory):
        return mesh_factory.__dict__["__factory_fields__"]
    else:
        return mesh_factory.__class__.__dict__["__factory_fields__"]



# -*- coding: utf-8 -*-

from PyR3.factory.fields.Unit import Length
from PyR3.factory.MeshFactory import MeshFactory
from PyR3.shortcut.edit import Edit
from PyR3.shortcut.mesh import addCircle
from PyR3.shortcut.transform import Transform


class CapacitorCase(MeshFactory):
    """
    Generates cylindrical meshes looking similar to electrolytic capacitor cases.
    """

    __author__ = "Krzysztof Wiśniewski"
    __version__ = "0.0.0"

    height = Length()
    radius = Length()

    def render(self) -> None:
        base_circle = addCircle(
            radius=self.radius,
        )
        with Edit(base_circle) as edit:
            edit.extrude()
            Transform.move((0, 0, 1))



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import io
import re
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext
import sys
import platform

from setuptools import setup, Distribution


def read(*names, **kwargs):
    with io.open(
        join(dirname(__file__), *names), encoding=kwargs.get("encoding", "utf8")
    ) as fh:
        return fh.read()


HAS_PLATNAME = any(["--plat-name" in string for string in sys.argv])
if HAS_PLATNAME:
    IS_LINUX = any(["manylinux1" in string for string in sys.argv])
    IS_WINDOWS = any(["win_amd64" in string for string in sys.argv])
else:
    IS_LINUX = platform.system() == "Linux"
    IS_WINDOWS = platform.system() == "Windows"


AUTHOR = "Krzysztof Wiśniewski"
EMAIL = "argmaster.world@gmail.com"
HOME_URL = "https://github.com/Argmaster/pyr3"

DESCRIPTION = (
    "A set of tools extending the capabilities of bpy (blender as a python module)."
)
LONG_DESCRIPTION = "%s\n%s" % (
    re.compile("^.. start-badges.*^.. end-badges", re.M | re.S).sub(
        "", read("README.rst")
    ),
    re.sub(":[a-z]+:`~?(.*?)`", r"``\1``", read("CHANGELOG.rst")),
)

CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: MIT License",
    "Operating System :: Unix",
    "Operating System :: POSIX",
    "Operating System :: Microsoft :: Windows",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: Implementation :: CPython",
    "Topic :: Utilities",
]

PROJECT_URLS = {
    "Documentation": "https://pyr3.readthedocs.io/",
    "Changelog": "https://pyr3.readthedocs.io/en/latest/changelog.html",
    "Issue Tracker": "https://github.com/Argmaster/pyr3/issues",
}


if __name__ == "__main__":
    with open("src/requirements.txt") as file:
        requirements = [r.strip() for r in file.readlines()]
    setup(
        name="PyR3",
        version="0.2.0",
        license="MIT",
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type="text/x-rst",
        author=AUTHOR,
        author_email=EMAIL,
        url=HOME_URL,
        packages=["PyR3"],
        package_dir={"": "src"},
        py_modules=[splitext(basename(path))[0] for path in glob("src/*.py")],
        include_package_data=True,
        zip_safe=False,
        classifiers=CLASSIFIERS,
        project_urls=PROJECT_URLS,
        keywords=[],
        python_requires=">=3.8",
        install_requires=requirements,
        extras_require={
            # eg:
            #   'rst': ['docutils>=0.11'],
            #   ':python_version=="2.6"': ['argparse'],
        }
    )



# -*- coding: utf-8 -*-
from __future__ import annotations

from types import SimpleNamespace

from PyR3.factory.MeshFactory import MeshFactory
from PyR3.factory.MeshFactory import getfields

from .Field import Field


class Struct(Field):
    """Parent class allowing to create custom struct classes grouping
    other field types by subclassing Struct in body of MeshFactory or
    another Strut field. Struct field value is a SimpleNamespace.

    """

    __instance: Struct = None

    def __new__(cls: Struct) -> Struct:
        if cls.__instance is None:
            fields = MeshFactory.get_field_names(cls.__qualname__, cls.__dict__)
            setattr(cls, "__factory_fields__", fields)
            cls.__del_fields(cls, fields)
            cls.__instance = super().__new__(cls)
        cls.__new__ = lambda _: cls.__instance
        return cls.__instance

    def __del_fields(cls, fields):
        for key in fields:
            delattr(cls, key)

    def __init__(self) -> None:
        pass

    def digest(self, params: dict = None) -> None:
        """Consumes dictionary of values and returns SimpleNamespace
        containing cleaned values of fields. Redundant params will be
        ignored. If a value is missing, None will be passed to coresponding
        field.

        :param params: dictionary of values, defaults to None
        :type params: dict, optional
        :return: namespace with cleaned values.
        :rtype: SimpleNamespace
        """
        namespace = SimpleNamespace()
        for name, field in getfields(self).items():
            param_value = params.get(name, None)
            cleaned_value = field.digest(param_value)
            setattr(namespace, name, cleaned_value)
        return namespace
